@@ -1,5 +1,7 @@
 from __future__ import annotations
 
+import time
+
 import pickle
 from pathlib import Path
 from functools import partial, wraps
@@ -7,13 +9,16 @@ from functools import partial, wraps
 import torchaudio
 from torchaudio import transforms as T
 from torchaudio.functional import resample
+from torchcodec.encoders import AudioEncoder
 
 import sounddevice as sd
 
+from loguru import logger
+
 import torch
 import torch.nn.functional as F
 from torch.fft import irfft
-from torch import nn, compiler, Tensor, tensor, is_tensor, cat, stft, hann_window
+from torch import nn, compiler, Tensor, tensor, is_tensor, cat, stft, hann_window, view_as_complex, view_as_real
 from torch.nn import LSTM, GRU, ConvTranspose1d, Module, ModuleList
 
 from numpy import ndarray
@@ -40,6 +45,14 @@ import matplotlib.pyplot as plt
 LSTM = partial(LSTM, batch_first = True)
 GRU = partial(GRU, batch_first = True)
 
+(
+    view_as_real,
+    view_as_complex
+) = tuple(compiler.disable()(fn) for fn in (
+    view_as_real,
+    view_as_complex
+))
+
 # functions
 
 def exists(v):
@@ -57,6 +70,9 @@ def identity(t):
 def is_empty(t: Tensor):
     return t.numel() == 0
 
+def current_time_ms():
+    return time.time() * 1000
+
 def round_down_to_multiple(num, mult):
     return (num // mult) * mult
 
@@ -64,6 +80,28 @@ def lens_to_mask(lens: Tensor, max_len):
     seq = torch.arange(max_len, device = lens.device)
     return einx.greater('b, n -> b n', lens, seq)
 
+# measure latency
+
+def decorate_print_latency(log_prefix = None):
+    def inner(fn):
+        @wraps(fn)
+        def decorated(*args, **kwargs):
+
+            start_time = current_time_ms()
+            out = fn(*args, **kwargs)
+            elapsed_ms = current_time_ms() - start_time
+
+            latency_msg = f'{int(elapsed_ms):.2f} ms'
+
+            if exists(log_prefix):
+                latency_msg = f'{log_prefix} {latency_msg}'
+
+            logger.debug(latency_msg)
+            return out
+
+        return decorated
+    return inner
+
 # residual
 
 def residual(fn):
@@ -199,6 +237,7 @@ class HSTasNet(Module):
         num_sources = 4,      # drums, bass, vocals, other
         torch_compile = False,
         use_gru = False,
+        spec_branch_use_phase = True,
         norm_before_mask_estimate = True # for some reason, training is unstable without a norm - improvise by adding an RMSNorm before final projection
     ):
         super().__init__()
@@ -241,7 +280,10 @@ class HSTasNet(Module):
             hop_length = overlap_len
         )
 
-        spec_dim_input = (n_fft // 2 + 1) * audio_channels
+        real_imag_dim = 2 if spec_branch_use_phase else 1
+        spec_dim_input = (n_fft // 2 + 1) * audio_channels * real_imag_dim
+
+        self.spec_branch_use_phase = spec_branch_use_phase # use the phase, proven out in another music sep paper
 
         self.spec_encode = nn.Sequential(
             Rearrange('(b s) f n ... -> b n (s f ...)', s = audio_channels),
@@ -251,7 +293,7 @@ class HSTasNet(Module):
         self.to_spec_masks = nn.Sequential(
             nn.RMSNorm(dim) if norm_before_mask_estimate else nn.Identity(),
             nn.Linear(dim, spec_dim_input * num_sources),
-            Rearrange('b n (s f t) -> (b s) f n t', s = audio_channels, t = num_sources)
+            Rearrange('b n (s f c t) -> (b s) f n c t', s = audio_channels, t = num_sources, c = real_imag_dim)
         )
 
         # waveform branch encoder
@@ -333,12 +375,21 @@ class HSTasNet(Module):
         if isinstance(audio, ndarray):
             audio = torch.from_numpy(audio)
 
+        if audio.ndim == 2:
+            # average stereo for now
+            audio = reduce(audio, 's n -> n', 'mean')
+
         assert audio.ndim == 1
         audio = audio.detach().cpu()
 
         # stft to magnitude to db
 
-        stft = torch.stft(audio, n_fft = self.n_fft, hop_length = self.hop_length, return_complex = True)
+        stft = T.Spectrogram(
+            n_fft = self.n_fft,
+            hop_length = self.hop_length,
+            power = 2.
+        )(audio)
+
         magnitude = stft.abs()
 
         db = T.AmplitudeToDB()(magnitude)
@@ -360,6 +411,7 @@ class HSTasNet(Module):
         plt.tight_layout()
 
         plt.savefig(str(save_path))
+        plt.close()
 
     # saving and loading
 
@@ -378,7 +430,8 @@ class HSTasNet(Module):
         if audio_tensor.ndim == 1:
             audio_tensor = repeat(audio_tensor, 'n -> s n', s = 2 if self.stereo else 1)
 
-        torchaudio.save(str(output_file), audio_tensor.cpu(), sample_rate = self.sample_rate)
+        encoder = AudioEncoder(audio_tensor.cpu(), sample_rate = self.sample_rate)
+        encoder.to_file(str(output_file))
 
         if verbose:
             print(f'audio saved to {str(output_file)}')
@@ -412,8 +465,7 @@ class HSTasNet(Module):
 
         config = pickle.loads(pkg['config'])
         model = cls(**config)
-
-        self.load(model, path, strict = strict)
+        model.load_state_dict(pkg['model'], strict = strict)
         return model
 
     # returns a function that closures the hiddens and past audio chunk for integration with sounddevice audio callback
@@ -422,7 +474,8 @@ class HSTasNet(Module):
         self,
         device = None,
         return_reduced_sources: list[int] | None = None,
-        auto_convert_to_stereo = True
+        auto_convert_to_stereo = True,
+        print_latency = False
     ):
         chunk_len = self.overlap_len
         self.eval()
@@ -479,6 +532,11 @@ class HSTasNet(Module):
 
             return transformed[..., -chunk_len:]
 
+        # print latency if needed
+
+        if print_latency:
+            fn = decorate_print_latency('stream chunk')(fn)
+
         return fn
 
     def sounddevice_stream(
@@ -488,6 +546,7 @@ class HSTasNet(Module):
         channels = None,
         device = None,
         auto_convert_to_stereo = True,
+        print_latency = False,
         **stream_kwargs
     ):
         assert len(return_reduced_sources) > 0
@@ -495,7 +554,8 @@ class HSTasNet(Module):
         transform_fn = self.init_stateful_transform_fn(
             return_reduced_sources = return_reduced_sources,
             device = device,
-            auto_convert_to_stereo = auto_convert_to_stereo
+            auto_convert_to_stereo = auto_convert_to_stereo,
+            print_latency = print_latency
         )
 
         # sounddevice stream callback, where raw audio can be transformed
@@ -660,7 +720,14 @@ class HSTasNet(Module):
 
         complex_spec, magnitude = self.stft(spec_audio_input)
 
-        spec = self.spec_encode(magnitude)
+        if self.spec_branch_use_phase:
+            real_imag_spec = torch.view_as_real(complex_spec)
+
+            spec = self.spec_encode(real_imag_spec)
+
+        else:
+
+            spec = self.spec_encode(magnitude)
 
         # handle encoding as detailed in original tasnet
         # to keep non-negative, they do a glu with relu on main branch
@@ -728,13 +795,22 @@ class HSTasNet(Module):
 
         spec_mask = self.to_spec_masks(spec)
 
-        magnitude, phase = complex_spec.abs(), complex_spec.angle()
+        if self.spec_branch_use_phase:
 
-        scaled_magnitude = multiply('b ..., b ... t -> (b t) ...', magnitude, spec_mask)
+            scaled_real_imag_spec = multiply('b ..., b ... t -> (b t) ...', real_imag_spec, spec_mask)
 
-        phase = repeat(phase, 'b ... -> (b t) ...', t = self.num_sources)
+            complex_spec_per_source = torch.view_as_complex(scaled_real_imag_spec.contiguous())
 
-        complex_spec_per_source = torch.polar(scaled_magnitude, phase)
+        else:
+            spec_mask = rearrange(spec_mask, '... 1 t -> ... t')
+
+            magnitude, phase = complex_spec.abs(), complex_spec.angle()
+
+            scaled_magnitude = multiply('b ..., b ... t -> (b t) ...', magnitude, spec_mask)
+
+            phase = repeat(phase, 'b ... -> (b t) ...', t = self.num_sources)
+
+            complex_spec_per_source = torch.polar(scaled_magnitude, phase)
 
         recon_audio_from_spec = self.stft.inverse(complex_spec_per_source)
 
