@@ -4,12 +4,16 @@ from functools import partial
 from random import random, randrange
 from pathlib import Path
 
+import torchaudio
+
+import musdb
+
 import torch
 from torch import cat, stack, tensor
 from torch.nn import Module
 from torch.optim import Adam
 from torch.optim.lr_scheduler import StepLR
-from torch.utils.data import Dataset, DataLoader, random_split
+from torch.utils.data import Dataset, DataLoader, ConcatDataset, random_split
 from torch.nn.utils.rnn import pad_sequence
 
 import numpy as np
@@ -63,7 +67,7 @@ def not_improved_last_n_steps(losses, steps):
 
     last_n_losses = losses[-(steps + 1):]
 
-    return (last_n_losses[1:] <= last_n_losses[:-1]).all().item()
+    return (last_n_losses[1:] > last_n_losses[:-1]).all().item()
 
 # dataset collation
 
@@ -80,6 +84,80 @@ def default_collate_fn(
 
 # dataset related
 
+# musdb18hq dataset wrapper
+
+class MusDB18HQ(Dataset):
+    def __init__(
+        self,
+        dataset_path: str | Path,
+        sep_filenames = ('drums', 'bass', 'vocals', 'other'),
+        max_audio_length_seconds = None
+    ):
+        if isinstance(dataset_path, str):
+            dataset_path = Path(dataset_path)
+
+        paths = []
+        mixture_paths = dataset_path.glob('**/*/mixture.wav')
+
+        # go through all the found paths to mixture.wav
+        # and make sure the separated wav files all exist
+
+        for mixture_path in mixture_paths:
+            parent_path = mixture_path.parent
+
+            if not all([(parent_path / f'{sep_filename}.wav').exists() for sep_filename in sep_filenames]):
+                continue
+
+            paths.append(parent_path)
+
+        self.paths = paths
+        self.sep_filenames = sep_filenames
+
+        # if the max_audio_length_seconds is set, will randomly sample a segment from the audio
+
+        assert not exists(max_audio_length_seconds) or max_audio_length_seconds > 0
+        self.max_audio_length_seconds = max_audio_length_seconds
+
+    def __len__(self):
+        return len(self.paths)
+
+    def __getitem__(self, idx):
+        path = self.paths[idx]
+
+        mixture_path = path / 'mixture.wav'
+
+        # get mixture as 'audio'
+
+        audio, sample_rate = torchaudio.load(str(mixture_path))
+
+        # get all separated audio with filenames as defined by `sep_filenames`
+
+        target_tensors = []
+        for sep_filename in self.sep_filenames:
+            sep_path = path / f'{sep_filename}.wav'
+            target, _ = torchaudio.load(str(sep_path))
+            target_tensors.append(target)
+
+        targets = stack(target_tensors)
+
+        # audio lengths for the uncompressed version is much longer
+
+        audio_length = audio.shape[-1]
+
+        if exists(self.max_audio_length_seconds):
+            max_length = self.max_audio_length_seconds * sample_rate
+
+            if audio_length > max_length:
+
+                start_index = randrange(audio_length - max_length)
+
+                audio = audio[..., start_index:(start_index + max_length)]
+                targets = targets[..., start_index:(start_index + max_length)]
+
+        # return
+
+        return audio, targets
+
 # wrap the musdb MultiTrack if detected
 
 class MusDBDataset(Dataset):
@@ -145,6 +223,31 @@ class StereoToMonoDataset(Dataset):
 
         return audio, targets
 
+class MaxSamples(Dataset):
+    def __init__(self, dataset: Dataset, max_samples):
+        self.dataset = dataset
+        self.max_samples = max_samples
+
+    def __len__(self):
+        return len(self.dataset)
+
+    def __getitem__(self, idx):
+        audio, targets = self.dataset[idx]
+
+        audio_length = audio.shape[-1]
+        max_length = self.max_samples
+
+        if audio_length <= max_length:
+            return audio, targets
+
+        start_index_range = audio_length - max_length
+        start_index = randrange(start_index_range)
+
+        audio = audio[..., start_index:(start_index + max_length)]
+        targets = targets[..., start_index:(start_index + max_length)]
+
+        return audio, targets
+
 # augmentations
 
 class GainAugmentation(Dataset):
@@ -264,8 +367,17 @@ class Trainer(Module):
     def __init__(
         self,
         model: HSTasNet,
-        dataset: Dataset | MusDB,
-        eval_dataset: Dataset | None = None,
+        dataset: (
+            Dataset |
+            list[Dataset | MusDB]
+            | MusDB
+            | None
+        ) = None,
+        concat_musdb_dataset = False,
+        use_full_musdb_dataset = False,
+        full_musdb_dataset_root = './data/musdb',
+        eval_dataset: Dataset | MusDB | None = None,
+        dataset_max_seconds = None,
         random_split_dataset_for_eval_frac = 0., # if set higher than 0., will split off this fraction of the training dataset for evaluation - eval_dataset must be None
         optim_klass = Adam,
         batch_size = 128,
@@ -333,11 +445,46 @@ class Trainer(Module):
             **optimizer_kwargs
         )
 
-        # dataset
+        # take care of datasets
+        # in the paper they use MusDB supplemented with own in-house dataset
+        # we will allow for automatic concatenation of MusDB (sample or full), or by itself if no dataset were given
 
-        if isinstance(dataset, MusDB):
-            # give musdb special treatment
-            dataset = MusDBDataset(dataset)
+        datasets = []
+
+        if exists(dataset):
+
+            # `dataset` can be a list already to be concatted together
+            # or just append to datasets list above
+
+            if isinstance(dataset, list):
+                datasets = dataset
+            else:
+                datasets.append(dataset)
+
+        if concat_musdb_dataset:
+            # concat the musdb dataset if need be
+
+            if use_full_musdb_dataset:
+                musdb_kwargs = dict(root = full_musdb_dataset_root)
+            else:
+                musdb_kwargs = dict(download = True)
+
+            musdb_dataset = musdb.DB(**musdb_kwargs)
+
+            datasets.append(musdb_dataset)
+
+        # convert MusDB dataset with wrapper if needed
+
+        datasets = [MusDBDataset(ds) if isinstance(ds, MusDB) else ds for ds in datasets] # wrap with musdb dataset to convert to (<audio>, <target>) pairs of right shape
+
+        if isinstance(eval_dataset, MusDB):
+            eval_dataset = MusDBDataset(eval_dataset)
+
+        # concat datasets
+
+        all_dataset = ConcatDataset(datasets)
+
+        assert len(all_dataset) > 0, 'your dataset is empty'
 
         # maybe split dataset for eval
 
@@ -345,17 +492,19 @@ class Trainer(Module):
         assert not (need_split_train_dataset and exists(eval_dataset)), f'`eval_dataset` must not be passed in if `random_split_dataset_for_eval_frac` set greater than 0.'
 
         if not exists(eval_dataset) and need_split_train_dataset:
-            train_size = int((1. - random_split_dataset_for_eval_frac) * len(dataset))
-            eval_size = len(dataset) - train_size
-            dataset, eval_dataset = random_split(dataset, (train_size, eval_size))
+            train_size = int((1. - random_split_dataset_for_eval_frac) * len(all_dataset))
+            eval_size = len(all_dataset) - train_size
+            train_dataset, eval_dataset = random_split(all_dataset, (train_size, eval_size))
+        else:
+            train_dataset = all_dataset
 
         # print
 
-        self.print(f'\ntraining on dataset of {len(dataset)} samples')
+        self.print(f'\ntraining on dataset of {len(train_dataset)} samples')
 
         # torch from this point on
 
-        dataset = CastTorch(dataset, device = device)
+        train_dataset = CastTorch(train_dataset, device = device)
 
         if exists(eval_dataset):
             eval_dataset = CastTorch(eval_dataset, device = device)
@@ -363,25 +512,34 @@ class Trainer(Module):
         # handle model is not stereo but data is stereo
 
         if not self.model_is_stereo:
-            dataset = StereoToMonoDataset(dataset)
+            train_dataset = StereoToMonoDataset(train_dataset)
 
             if exists(eval_dataset):
                 eval_dataset = StereoToMonoDataset(eval_dataset)
 
         # augmentations
-        # only for training `dataset`
+        # only for training `train_dataset`
 
         if augment_gain:
-            dataset = GainAugmentation(dataset)
+            train_dataset = GainAugmentation(train_dataset)
 
         collate_fn = default_collate_fn
 
         if augment_remix:
             collate_fn = compose(collate_fn, partial(augment_remix_fn, frac_augment = augment_remix_frac))
 
+        # random sample segments if `dataset_max_seconds` is set
+
+        if exists(dataset_max_seconds):
+            max_samples = self.sample_rate * dataset_max_seconds
+            train_dataset = MaxSamples(train_dataset, max_samples)
+
+            if exists(eval_dataset):
+                eval_dataset = MaxSamples(eval_dataset, max_samples)
+
         # dataloader
 
-        dataloader = DataLoader(dataset, batch_size = batch_size, drop_last = True, shuffle = True, collate_fn = collate_fn)
+        dataloader = DataLoader(train_dataset, batch_size = batch_size, drop_last = True, shuffle = True, collate_fn = collate_fn)
 
         eval_dataloader = None
 
@@ -506,6 +664,16 @@ class Trainer(Module):
         exceeds_max_step = False
         past_eval_losses = [] # for learning rate decay and early stopping detection
 
+        # the function for saving checkpoints
+
+        def save_checkpoints(checkpoint_index):
+            self.unwrapped_model.save(self.checkpoint_folder / f'hs-tasnet.ckpt.{checkpoint_index}.pt')
+
+            if self.use_ema:
+                self.ema_model.ema_model.save(self.checkpoint_folder /f'hs_tasnet.ema.ckpt.{checkpoint_index}.pt') # save ema
+
+        # go through all epochs
+
         for epoch in range(1, self.max_epochs + 1):
 
             self.model.train()
@@ -585,7 +753,7 @@ class Trainer(Module):
                     # take a random sample from eval dataset and store the audio and spectrogram results
 
                     rand_index = randrange(len(self.eval_dataset))
-                    eval_audio, *_ = self.eval_dataset[rand_index]
+                    eval_audio, eval_targets = self.eval_dataset[rand_index]
 
                     with torch.no_grad():
                         model.eval()
@@ -599,25 +767,45 @@ class Trainer(Module):
                     one_eval_result_folder = self.eval_results_folder / str(epoch)
                     one_eval_result_folder.mkdir(parents = True, exist_ok = True)
 
-                    # save files to folder
+                    # save eval files (audio and spectrogram) to folder
 
                     eval_audio_paths = []
+                    eval_spec_img_paths = []
 
                     model.save_tensor_to_file(one_eval_result_folder / 'audio.mp3', eval_audio, overwrite = True)
+                    model.save_spectrogram_figure(one_eval_result_folder / 'spec.png', eval_audio, overwrite = True)
 
                     eval_audio_paths.append(('eval_audio', str(one_eval_result_folder / 'audio.mp3')))
+                    eval_spec_img_paths.append(('eval_spec', str(one_eval_result_folder / 'spec.png')))
 
-                    for index, target_audio in enumerate(separated_audio):
-                        model.save_tensor_to_file(one_eval_result_folder / f'target.separated.{index}.mp3', target_audio, overwrite = True)
+                    for index, (pred_target_audio, target_audio) in enumerate(zip(separated_audio, eval_targets)):
 
-                        eval_audio_paths.append((f'target_audio_{index}', str(one_eval_result_folder / f'{index}.mp3')))
+                        saved_audio_path = one_eval_result_folder / f'target.separated.{index}.mp3'
+                        pred_saved_audio_path = one_eval_result_folder / f'pred.target.separated.{index}.mp3'
+
+                        saved_spectrogram_path = one_eval_result_folder / f'target.separated.spectrogram.{index}.png'
+                        pred_saved_spectrogram_path = one_eval_result_folder / f'pred.target.separated.spectrogram.{index}.png'
+
+                        model.save_tensor_to_file(saved_audio_path, target_audio, overwrite = True)
+                        model.save_tensor_to_file(pred_saved_audio_path, pred_target_audio, overwrite = True)
+
+                        model.save_spectrogram_figure(saved_spectrogram_path, target_audio, overwrite = True)
+                        model.save_spectrogram_figure(pred_saved_spectrogram_path, pred_target_audio, overwrite = True)
+
+                        eval_audio_paths.append((f'target_audio_{index}', str(pred_saved_audio_path)))
+                        eval_spec_img_paths.append((f'target_spec_{index}', str(pred_saved_spectrogram_path)))
+
+                    # log audio and spec images to wandb experiment if need be
 
                     if self.use_wandb:
-                        from wandb import Audio
+                        from wandb import Audio, Image
 
                         for name, eval_audio_path in eval_audio_paths:
                             eval_logs[name] = Audio(eval_audio_path, sample_rate = self.sample_rate)
 
+                        for name, eval_spec_img_path in eval_spec_img_paths:
+                            eval_logs[name] = Image(eval_spec_img_path)
+
                 self.log(**eval_logs)
 
             # maybe save
@@ -625,14 +813,12 @@ class Trainer(Module):
             self.wait()
 
             if (
-                divisible_by(epoch + 1, self.checkpoint_every) and
+                divisible_by(epoch, self.checkpoint_every) and
                 self.is_main
             ):
-                checkpoint_index = (epoch + 1) // self.checkpoint_every
-                self.unwrapped_model.save(self.checkpoint_folder / f'hs-tasnet.ckpt.{checkpoint_index}.pt')
+                checkpoint_index = epoch // self.checkpoint_every
 
-                if self.use_ema:
-                    self.ema_model.ema_model.save(self.checkpoint_folder /f'hs_tasnet.ema.ckpt.{checkpoint_index}.pt') # save ema
+                save_checkpoints(checkpoint_index)
 
             # determine lr decay and early stopping based on eval
 
@@ -644,12 +830,13 @@ class Trainer(Module):
                 # decay lr if criteria met
 
                 if not_improved_last_n_steps(last_n_eval_losses, self.decay_lr_if_not_improved_steps):
+                    self.print(f'decaying learning rate as validation has not improved for {self.decay_lr_if_not_improved_steps} steps')
                     self.scheduler.step()
 
                 # early stop if criteria met
 
                 if not_improved_last_n_steps(last_n_eval_losses, self.early_stop_steps):
-                    self.print(f'early stopping at epoch {epoch} since last three eval losses have not improved: {last_n_eval_losses}')
+                    self.print(f'early stopping at epoch {epoch} since last three eval losses have not improved: {last_n_eval_losses.tolist()}')
                     break
 
             if exceeds_max_step:
@@ -657,6 +844,8 @@ class Trainer(Module):
 
         # cleanup accelerator
 
+        save_checkpoints(-1) # save one last time with checkpoint index -1
+
         self.accelerator.end_training()
 
         self.print(f'training complete')
