@@ -9,11 +9,15 @@ import torch
 @param('stereo', (False, True))
 @param('use_gru', (False, True))
 @param('var_audio_lens', (False, True))
+@param('spec_branch_use_phase', (False, True))
+@param('torch_compile', (False, True))
 def test_model(
     small,
     stereo,
     use_gru,
-    var_audio_lens
+    var_audio_lens,
+    spec_branch_use_phase,
+    torch_compile
 ):
     from hs_tasnet.hs_tasnet import HSTasNet
 
@@ -21,7 +25,9 @@ def test_model(
         dim = 512,
         small = small,
         stereo = stereo,
-        use_gru = use_gru
+        use_gru = use_gru,
+        spec_branch_use_phase = spec_branch_use_phase,
+        torch_compile = torch_compile
     )
 
     shape = (2, 1017 * 12) if stereo else (1017 * 12,)
@@ -51,9 +57,11 @@ def test_model(
 
 @param('with_eval', (False, True))
 @param('with_ema', (False, True))
+@param('list_dataset', (False, True))
 def test_trainer(
     with_eval,
-    with_ema
+    with_ema,
+    list_dataset
 ):
     from hs_tasnet.hs_tasnet import HSTasNet
     from hs_tasnet.trainer import Trainer
@@ -80,9 +88,14 @@ def test_trainer(
             targets = torch.rand(4, 1024 * 10)
             return audio, targets
 
+    if list_dataset:
+        dataset = [MusicSepDataset(), MusicSepDataset()]
+    else:
+        dataset = MusicSepDataset()
+
     trainer = Trainer(
         model,
-        dataset = MusicSepDataset(),
+        dataset = dataset,
         eval_dataset = EvalMusicSepDataset() if with_eval else None,
         batch_size = 4,
         max_epochs = 3,
@@ -100,6 +113,17 @@ def test_trainer(
         assert eval_folder.exists() and eval_folder.is_dir()
 
 def test_audio_processing():
+    import sounddevice as sd
+
+    device_list = sd.query_devices()
+    num_devices = len(device_list)
+
+    if num_devices == 0:
+        pytest.skip()
+        return
+
+    # test sound device stream
+
     from hs_tasnet.hs_tasnet import HSTasNet
 
     model = HSTasNet(
